/*
	PROforma, PROgs FOnt & Raster MAnager

	Routines to construct and draw a path

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 July 1993
	09.11.1995 removed all references to PF_PATH_FAST
	24.06.1996 - PF_PATH_STROKED now has to be converted by PFPathMethod and
				 PFLineWidth operators
			   - thick lines are now drawn in device coordinates (faster)
			   - thick lines are now normalized for uniform linewidth
	26.06.1996 - thick lines now also ok when pixels not square
	18.09.1996 PFP routines were no longer correct when PageOrigin!=(0,0)
	23.10.1996 - Curves adjusted for PageOrigin twice - fixed
			   - stoked lines also performed PFPathClose when drawing (?)
	25.11.1996 added high level drawing commands
	15.02.1997 PFPie should now function properly
	16.02.1996 PFPArcR should now function properly (this caused bugs is PFArcR
			   and PFPie as well)
*/

#include <math.h>

#include "gstate.h"
#include "system.h"
#include "path.h"

/* ACTUAL DRAWING COMMAND */
#define tline(x1,y1,x2,y2) addflatline(gstate, x1,y1,x2,y2)

#define tcurve(x0,y0,x1,y1,x2,y2,x3,y3) addflatcurve(gstate, \
	x0,y0,x1,y1,x2,y2,x3,y3,gstate->flatness, addflatline, (Bitmap)gstate)


Error PFPathDraw(Gstate gstate)
{
	/* now this always boils down to filling a path unless the */
	/* line which is hairline, then nothing has to be done */

	switch (gstate->pathtype)
	{
	case PF_PATH_FILLED:
		PFPathClose(gstate);
	case PF_PATH_THICKLINE:
		fillpath(gstate, gstate->raster->bitdraw->TrapezoidHorizontal,
				 gstate->raster, PF_WINDING_RULE);
		break;
	case PF_PATH_EOFILLED:
		PFPathClose(gstate);
		fillpath(gstate, gstate->raster->bitdraw->TrapezoidHorizontal,
				 gstate->raster, PF_EVEN_ODD_RULE);
		break;
	}
	return PFPathClear(gstate);
}


/* path construction routines, these either construct the path, or draw it */
/* immediatly if pathtype==PF_PATH_HAIRLINE */


/* ABSOLUTE COMMANDS */

static Error _PFPMoveTo(Gstate gstate, pt x, pt y)
{
	coor c;

	c.x=x; c.y=y;

	/* close previous subpath if filled */
	if (gstate->pathtype==PF_PATH_FILLED || gstate->pathtype==PF_PATH_EOFILLED)
		PFPathClose(gstate);

	gstate->subpathorg=gstate->posdev=c;
	return ERR_OK;
}

Error PFPMoveTo(Gstate gstate, pt x, pt y)
{
	return _PFPMoveTo(gstate,x-gstate->Origin.x,y-gstate->Origin.y);
}

Error PFMoveTo(Gstate gstate, pt x, pt y)
{
	coor c;

	c.x=x; c.y=y;
	user2dev(gstate, &c);
	return _PFPMoveTo(gstate, c.x,c.y);
}

static Error _PFPLineTo(Gstate gstate, pt x2, pt y2)
{
	pt x1=gstate->posdev.x;
	pt y1=gstate->posdev.y;

	gstate->posdev.x=x2;
	gstate->posdev.y=y2;

	switch (gstate->pathtype)
	{
	case PF_PATH_FILLED: case PF_PATH_EOFILLED:
		tline(x1,y1,x2,y2);
		break;

	case PF_PATH_HAIRLINE:
		gstate->raster->bitdraw->HairLine(gstate->raster, x1,y1, x2,y2);
		break;

	case PF_PATH_THICKLINE:
		{
			pt dx,dy,cap,xcap,ycap,xext,yext;
			pt xycap, yxcap, xyext, yxext;
			pt len, ddx, ddy;

			if (gstate->linexmid)
				x1=pt_center(x1), x2=pt_center(x2);
			else
				x1=pt_round(x1), x2=pt_round(x2);
			if (gstate->lineymid)
				y1=pt_center(y1), y2=pt_center(y2);
			else
				y1=pt_round(y1), y2=pt_round(y2);

			dx=x2-x1; dy=y2-y1;
			ddx=fixmul(dx,gstate->linexppp);
			ddy=fixmul(dy,gstate->lineyppp);
			len=fixdist(ddx,ddy);
			if (len==0) break;
			cap=fixdiv(gstate->linewidth,len);
			xcap=xycap=fixmul(dy,cap); xext=xyext=fixmul(xcap,ptmagic);
			ycap=yxcap=fixmul(dx,cap); yext=yxext=fixmul(ycap,ptmagic);
			if (!gstate->uniformgrid)
			{
				pt y2x=gstate->liney2x, x2y=gstate->linex2y;
				xcap=fixmul(xcap,y2x);
				xext=fixmul(xext,y2x);
				ycap=fixmul(ycap,x2y);
				yext=fixmul(yext,x2y);
			}

			/* now start drawing the line with round cap */
			tline(x1-xcap, y1+ycap, x2-xcap, y2+ycap);
			tline(x2+xcap, y2-ycap, x1+xcap, y1-ycap);
			tcurve(x2-xcap, y2+ycap, x2-xcap+yxext, y2+ycap+xyext,
						x2+yxcap-xext, y2+xycap+yext, x2+yxcap, y2+xycap);
			tcurve(x2+yxcap, y2+xycap, x2+yxcap+xext, y2+xycap-yext,
						x2+xcap+yxext, y2-ycap+xyext, x2+xcap, y2-ycap);
			tcurve(x1+xcap, y1-ycap, x1+xcap-yxext, y1-ycap-xyext,
						x1-yxcap+xext, y1-xycap-yext, x1-yxcap, y1-xycap);
			tcurve(x1-yxcap, y1-xycap, x1-yxcap-xext, y1-xycap+yext,
						x1-xcap-yxext, y1+ycap-xyext, x1-xcap, y1+ycap);
		}
		break;
	}
	return ERR_OK;
}

Error PFPLineTo(Gstate gstate, pt x, pt y)
{
	return _PFPLineTo(gstate,x-gstate->Origin.x,y-gstate->Origin.y);
}

Error PFLineTo(Gstate gstate, pt x, pt y)
{
	coor c;

	c.x=x; c.y=y;
	user2dev(gstate, &c);
	return _PFPLineTo(gstate, c.x,c.y);
}

Error PFPathClose(Gstate gstate)
{
	return _PFPLineTo(gstate, gstate->subpathorg.x, gstate->subpathorg.y);
}
 
static Error FlatCurve(Gstate gstate, pt fx, pt fy, pt tx, pt ty)
{
	return _PFPLineTo(gstate,tx,ty);
}

#define _PFPCurveTo(gstate, cx1, cy1, cx2, cy2, x2, y2) \
	addflatcurve(gstate, gstate->posdev.x,gstate->posdev.y, cx1,cy1, cx2,cy2, \
				x2,y2, gstate->flatness, FlatCurve, (Bitmap)gstate);

Error PFPCurveTo(Gstate gstate, pt cx1, pt cy1, pt cx2, pt cy2, pt x2, pt y2)
{
	pt x=gstate->Origin.x;
	pt y=gstate->Origin.y;
	return _PFPCurveTo(gstate,cx1-x,cy1-y,cx2-x,cy2-y,x2-x,y2-y);
}

Error PFCurveTo(Gstate gstate, pt cx1, pt cy1, pt cx2, pt cy2, pt x, pt y)
{
	coor c,c1,c2;

	c.x=x; c.y=y; c1.x=cx1; c1.y=cy1; c2.x=cx2; c2.y=cy2;

	/* convert to device coordinates */
	user2dev(gstate, &c);
	user2dev(gstate, &c1);
	user2dev(gstate, &c2);

	return _PFPCurveTo(gstate, c1.x,c1.y, c2.x,c2.y, c.x,c.y);
}

/* RELATIVE COMMANDS */

Error PFMoveR(Gstate gstate, pt x, pt y)
{
	coor c;
	c.x=x; c.y=y;
	reluser2dev(gstate, &c);
	return _PFPMoveTo(gstate, c.x+gstate->posdev.x, c.y+gstate->posdev.y);
}

Error PFLineR(Gstate gstate, pt x, pt y)
{
	coor c;
	c.x=x; c.y=y;
	reluser2dev(gstate, &c);
	return _PFPLineTo(gstate, c.x+gstate->posdev.x, c.y+gstate->posdev.y);
}

Error PFCurveR(Gstate gstate, pt cx1, pt cy1, pt cx2, pt cy2, pt x, pt y)
{
	pt x0=gstate->posdev.x;
	pt y0=gstate->posdev.y;
	coor c1,c2,c;

	c.x=x; c.y=y; c1.x=cx1; c1.y=cy1; c2.x=cx2; c2.y=cy2;
	reluser2dev(gstate, &c);
	reluser2dev(gstate, &c1);
	reluser2dev(gstate, &c2);

	return _PFPCurveTo(gstate, c1.x+x0, c1.y+y0, c2.x+x0, c2.y+y0, c.x+x0, c.y+y0);
}

Error PFPMoveR(Gstate gstate, pt x, pt y)
{
	return _PFPMoveTo(gstate, x+gstate->posdev.x, y+gstate->posdev.y);
}

Error PFPLineR(Gstate gstate, pt x, pt y)
{
	return _PFPLineTo(gstate, x+gstate->posdev.x, y+gstate->posdev.y);
}

Error PFPCurveR(Gstate gstate, pt x1, pt y1, pt x2, pt y2, pt x3, pt y3)
{
	pt x0=gstate->posdev.x;
	pt y0=gstate->posdev.y;
	return _PFPCurveTo(gstate, x1+x0, y1+y0, x2+x0, y2+y0, x3+x0, y3+y0);
}

/* ----- high level drawing routines ----- */

/* PFArcTo : the first coor pair is the point where the two tangents cross */

Error PFPArcTo(Gstate gstate, pt tx, pt ty, pt x3, pt y3)
{
	pt x0=gstate->posdev.x;
	pt y0=gstate->posdev.y;

	return _PFPCurveTo(gstate,x0+fixmul(ptmagic,tx-x0),y0+fixmul(ptmagic,ty-y0),
							  x3+fixmul(ptmagic,tx-x3),y3+fixmul(ptmagic,ty-y3),
							  x3,y3);
}

Error PFPArcR(Gstate gstate, pt tx, pt ty, pt x3, pt y3)
{
	pt x0=gstate->posdev.x;
	pt y0=gstate->posdev.y;
	pt ex=x0+x3, ey=y0+y3;

	return _PFPCurveTo(gstate,x0+fixmul(ptmagic,tx	 ),y0+fixmul(ptmagic,ty   ),
							  ex+fixmul(ptmagic,tx-x3),ey+fixmul(ptmagic,ty-y3),
							  ex,ey);
}

Error PFArcTo(Gstate gstate, pt tx, pt ty, pt ex, pt ey)
{
	coor tc, ec;
	tc.x=tx; tc.y=ty;
	ec.x=ex; ec.y=ey;
	user2dev(gstate, &tc);
	user2dev(gstate, &ec);
	return PFPArcTo(gstate, tc.x, tc.y, ec.x, ec.y);
}

Error PFArcR(Gstate gstate, pt tx, pt ty, pt ex, pt ey)
{
	coor tc, ec;
	tc.x=tx; tc.y=ty;
	ec.x=ex; ec.y=ey;
	reluser2dev(gstate, &tc);
	reluser2dev(gstate, &ec);
	return PFPArcR(gstate, tc.x, tc.y, ec.x, ec.y);
}

Error PFCircle(Gstate gstate, pt r)
{
	pt magicr=fixmul(ptmagic,r);
	pt cigamr=fixmul(ptcigam,r);

	/* just draw a circle */
	PFMoveR(gstate,0,-r);
/*	  PFLineR(gstate, r,r);   */
/*	  PFLineR(gstate, -r,r);  */
/*	  PFLineR(gstate, -r,-r); */
/*	  PFLineR(gstate, r,-r);  */
	PFCurveR(gstate, magicr,0, r,cigamr, r,r);
	PFCurveR(gstate, 0,magicr, -cigamr,r, -r,r);
	PFCurveR(gstate, -magicr,0, -r,-cigamr, -r,-r);
	PFCurveR(gstate, 0,-magicr, cigamr,-r, r,-r);
	return ERR_OK;
}

Error PFPie(Gstate gstate, pt r, pt sdcos, pt sdsin,
						   pt edcos, pt edsin, char closed)
{
	/* swap sinuses around */
	sdsin = -sdsin; edsin = -edsin;

	/* pt 500 is about 0.0076 */
	if ( (abs(sdcos - edcos) < 500) && (abs(sdsin - edsin) < 500) )
		PFCircle(gstate,r);
	else
	{
		/* we have to draw an arc */
		pt x0, y0, sin0, cos0, cosdf, sindf, trad;

		sin0=fixmul(r, sdsin);
		cos0=fixmul(r, sdcos);
		/* the start-point */
		x0=cos0;
		y0=sin0;
		if (closed)
			PFLineR(gstate,x0,y0);
		else
			PFMoveR(gstate,x0,y0);
		/* quadrant reduction -- we can only draw arcs that are smaller */
		/* than 90 degrees apart with the PFArc routines */
		cosdf = fixmul(sdcos, edcos) + fixmul(sdsin, edsin);
		sindf = fixmul(edsin, sdcos) - fixmul(sdsin, edcos);
		/* while the difference is bigger than 90 degrees */
		while (cosdf < 0  || sindf < 0)
		{
			pt w = cos0; cos0 = -sin0; sin0 = w;
			w = sdcos; sdcos = -sdsin; sdsin = w;
			PFArcR(gstate, cos0, sin0, cos0-x0, sin0-y0);
			x0 = cos0; y0 = sin0;
			cosdf = fixmul(sdcos, edcos) + fixmul(sdsin, edsin);
			sindf = fixmul(edsin, sdcos) - fixmul(sdsin, edcos);
		}
		/* now we have an angle < 90 deg */
		/* Compute the intersection of the tangents. */

		/* we only need the half angle */
		sindf = fixsqrt((pt_one - cosdf) / 2);
		cosdf = fixsqrt((pt_one + cosdf) / 2);

		if (cosdf)
		{
			trad = fixdiv(sindf,cosdf);
			PFArcR(gstate, -fixmul(trad, sin0),
							fixmul(trad, cos0),
						   -x0 + fixmul(r, edcos),
						   -y0 + fixmul(r, edsin));
		}
		if (closed) PFPathClose(gstate);
	}
	return ERR_OK;
}

Error PFRectangle(Gstate gstate, pt a, pt b, pt r)
{
	if (a<0)
	{
		PFMoveR(gstate,a,0);
		a=-a;
	}
	if (b<0)
	{
		PFMoveR(gstate,0,b);
		b=-b;
	}
	if (r>a/2) r=a/2;
	if (r>b/2) r=b/2;
	if (r == 0)
	{
		PFLineR(gstate,a,0);
		PFLineR(gstate,0,b);
		PFLineR(gstate,-a,0);
		PFPathClose(gstate);
	}
	else
	{
		pt magicr=fixmul(ptmagic,r);
		pt cigamr=fixmul(ptcigam,r);
		pt aside=a-2*r, bside=b-2*r;

		PFMoveR(gstate,0,r);
		PFCurveR(gstate, 0,-magicr, cigamr,-r, r,-r);
		PFLineR(gstate,aside,0);
		PFCurveR(gstate, magicr,0, r,cigamr, r,r);
		PFLineR(gstate,0,bside);
		PFCurveR(gstate, 0,magicr, -cigamr,r, -r,r);
		PFLineR(gstate,-aside,0);
		PFCurveR(gstate, -magicr,0, -r,-cigamr, -r,-r);
		PFPathClose(gstate);
	}
	return ERR_OK;
}

Error PFPRectangle(Gstate gstate, pt a, pt b)
{
	PFPLineR(gstate,a,0);
	PFPLineR(gstate,0,b);
	PFPLineR(gstate,-a,0);
	return PFPathClose(gstate);
}
